"""This module describes a result for a query or bulk query"""
from rqdb.errors import DBError as Error
from typing import Any, Iterator, List, Optional


class ResultItem:
    """The result of a single query, either a standalone query or
    a member of a bulk query.
    """

    def __init__(
        self,
        results: Optional[List[list]] = None,
        last_insert_id: Optional[int] = None,
        rows_affected: Optional[int] = None,
        error: Optional[str] = None,
    ):
        self.results = results
        """The rows returned by the query. This is the "values" part of the result
        item returned by the RQLite API. This is the ordered list of values that
        were selected in this row, so, e.g.,

        "SELECT 1, 2, 3"

        would lead to the results list being [1, 2, 3].

        This is None if either the query was not a SELECT query or if the query
        failed.
        """

        self.last_insert_id = last_insert_id
        """The ID of the last inserted row after this query. This is None if
        the query was a SELECT query or if the query failed. This is typically
        only meaningful for INSERT queries.

        Note that the value of the last insert id may change arbitrarily between
        non-insert queries due to request interleaving on the underlying cursor
        on the actual rqlite nodes.
        """

        self.rows_affected = rows_affected
        """The number of rows affected by this query. This is None if the query
        was a SELECT query or if the query failed. This is typically only
        meaningful for INSERT, UPDATE, and DELETE queries.
        """

        self.error = error
        """The actual error text returned by the RQLite server. This is None if
        the query succeeded. It is helpful to use the Error object to determine
        the type of error.
        """

    @property
    def rowcount(self) -> int:
        """The number of rows returned by this query. 0 if the query was
        not a select query or if the query failed.
        """
        if self.results is None:
            return 0
        return len(self.results)

    def raise_on_error(self, hint=None) -> "ResultItem":
        """Raises an error if this result item has an error. Otherwise,
        returns self.

        Args:
            hint (str): If specified, used as part of the error message to
                provide additional context.

        Returns:
            self

        Raises:
            Error: If this result item has an error.
        """
        if self.error is not None:
            if hint is None:
                raise Error(self.error)
            raise Error(f"{hint}: {self.error}", raw_message=self.error)

        return self

    def cursor(self) -> "ResultItemCursor":
        """Returns a cursor-like object that allows paginating over the results
        using fetchone() style methods. Raises an error if this is not the result
        of a successful SELECT query.
        """
        if self.results is None:
            raise Error("Cannot get cursor for non-SELECT query")
        return ResultItemCursor(self.results)

    def __repr__(self) -> str:
        return f"ResultItem(results={repr(self.results)}, last_insert_id={repr(self.last_insert_id)}, rows_affected={repr(self.rows_affected)}, error={repr(self.error)})"

    @classmethod
    def parse(cls, result: dict):
        """Parses a result from the RQLite API into a ResultItem object."""
        if "error" in result:
            return ResultItem(error=result["error"])

        if "values" in result:
            return ResultItem(results=result["values"])

        return ResultItem(
            last_insert_id=result.get("last_insert_id"),
            rows_affected=result.get("rows_affected"),
        )


class ResultItemCursor:
    """Describes a cursor-like object for a result item, which allows paginating
    the values with the familiar fetchone(), fetchmany(), and fetchall() methods.
    This only applies to successful SELECT queries.
    """

    def __init__(self, results: List[tuple]):
        self.results = results
        """The list of result items returned by the query"""

        self.index = 0
        """The index of the current result item"""

    def fetchone(self) -> Optional[List[Any]]:
        """Returns the next row in the result set. Returns None if there are no
        more rows.
        """
        if self.index >= len(self.results):
            return None
        result = self.results[self.index]
        self.index += 1
        return result

    def fetchmany(self, size: int = None) -> List[List[Any]]:
        """Returns the next `size` rows in the result set. Returns an empty
        list if there are no more rows.
        """
        if self.index >= len(self.results):
            return []
        if size is None:
            size = len(self.results) - self.index
        result = self.results[self.index : self.index + size]
        self.index += len(result)
        return result

    def fetchall(self) -> List[List[Any]]:
        """Returns all remaining rows in the result set. Returns an empty
        list if there are no more rows.
        """
        if self.index >= len(self.results):
            return []
        result = self.results[self.index :]
        self.index = len(self.results)
        return result

    @property
    def rowcount(self) -> int:
        """Returns the total number of rows in the result set."""
        return len(self.results)


class BulkResult:
    """Describes the result of a bulk query. This is the result of many
    individual queries, potentially in the same transaction.

    This can be indexed by the index of the query in the bulk query.
    """

    def __init__(self, items: List[ResultItem]) -> None:
        self.items = items
        """The individual result items for the bulk query."""

    def raise_if_error_before(self, idx: int) -> "BulkResult":
        """Raises an error if any queries before the given index have errors.
        Otherwise, returns self.

        Args:
            idx (int): The index of the query to check for errors.

        Returns:
            self

        Raises:
            Error: If any queries before the given index have errors.
        """
        if idx < 0:
            raise ValueError("idx must be >= 0")

        for i in range(min(len(self.items), idx)):
            self.items[i].raise_on_error(f"query idx={idx}")

        return self

    def raise_on_error(self, hint: Optional[str] = None) -> "BulkResult":
        """Raises an error if any of the queries in this bulk query have errors.
        Otherwise, returns self.

        Returns:
            self

        Raises:
            Error: If any of the queries in this bulk query have errors.
        """
        hint = "" if hint is None else f"; {hint}"
        for idx, item in enumerate(self.items):
            item.raise_on_error(f"query idx={idx}{hint}")

        return self

    def __getitem__(self, idx: int) -> ResultItem:
        return self.items[idx]

    def __len__(self) -> int:
        return len(self.items)

    def __iter__(self) -> Iterator[ResultItem]:
        return iter(self.items)

    def __repr__(self) -> str:
        return f"BulkResult(items={repr(self.items)})"

    def __str__(self) -> str:
        return str(self.items)

    @classmethod
    def parse(cls, payload: dict) -> "BulkResult":
        """Parses a result from the RQLite API into a BulkResult object."""
        return BulkResult(
            [ResultItem.parse(item) for item in payload.get("results", [])]
        )



from typing import List, Optional, Tuple


class DBError(Exception):
    """Describes an error returned by rqlite"""

    def __init__(self, msg: str, raw_message: Optional[str] = None) -> None:
        super().__init__(msg)
        self.message = msg
        """Our text for the error"""

        self.raw_message = raw_message or msg
        """The raw message returned by rqlite"""

    @property
    def is_foreign_key(self) -> bool:
        """True if this error is a foreign key error, false otherwise"""
        return self.raw_message == "FOREIGN KEY constraint failed"

    @property
    def is_unique(self) -> bool:
        """True if this error is a unique constraint error, false otherwise"""
        return "unique" in self.raw_message.lower()

    @property
    def is_syntax(self) -> bool:
        """True if this error is a syntax error, false otherwise"""
        return self.raw_message.endswith("syntax error")

    @property
    def is_stale(self) -> bool:
        """True if this error indicates a stale read occurred, false otherwise"""
        return self.raw_message == "stale read"


class ConnectError(Exception):
    """Describes a timeout or other failure when connecting to a rqlite node"""

    def __init__(self, msg: str, host: str) -> None:
        super().__init__(msg)
        self.host = host
        """The host we tried to connect to"""


class UnexpectedResponse(Exception):
    """Describes an unexpected response from rqlite"""

    def __init__(self, host: str, msg: str) -> None:
        super().__init__(msg)
        self.host = host
        """The host which returned the unexpected response"""


class MaxAttemptsError(Exception):
    """Describes exhausting all attempts on all nodes in a rqlite cluster"""

    def __init__(self, node_path: List[Tuple[str, Exception]]) -> None:
        super().__init__(f"Exhausted all attempts on all nodes; {node_path=}")
        self.node_path = node_path
        """A list of (host, error) tuples for each node we tried to connect to"""


class MaxRedirectsError(Exception):
    """Describes too many redirects when connecting to a rqlite node"""

    def __init__(self, host: str, redirect_path: List[str]) -> None:
        super().__init__(f"Too many redirects connecting to {host}: {redirect_path}")
        self.host = host
        """The host we tried to connect to"""

        self.redirect_path = redirect_path
        """The path we followed to get to the final host"""



"""This module is for any necessary SQL preprocessing required for the
rqlite client.
"""
import re
from typing import Tuple


WITH_MATCHER = re.compile(
    r"WITH( RECURSIVE)?\s+(,?\s*\S+\s+AS\s+\(.+?\))+\s+(?P<cmd>INSERT|UPDATE|DELETE|SELECT)",
    re.IGNORECASE,
)
"""The matcher to use for determing the sql command for a SQL string with a WITH clause"""


def get_sql_command(sql_str: str) -> str:
    """Determines which sql command is being used in the given SQL string.

    Args:
        sql_str (str): The SQL string to parse.

    Returns:
        The corresponding command (SELECT, INSERT, etc.)

    Raises:
        Exception: If the command could not be determined
    """
    sql_str = sql_str.lstrip()
    if sql_str[:4].upper() == "WITH":
        match = WITH_MATCHER.match(sql_str)
        if match is None:
            raise Exception(f"could not determine SQL command for {sql_str=}")
        return match.group("cmd").upper()

    whitespace_idx = -1
    for i, c in enumerate(sql_str):
        if c.isspace():
            whitespace_idx = i
            break
    else:
        raise Exception(f"could not determine SQL command for {sql_str=}")
    return sql_str[:whitespace_idx].upper()


def clean_nulls(sql_str: str, args: tuple) -> Tuple[str, tuple]:
    """Currently RQLite does not handle NULL-arguments. We have to
    do our best to manipulate the SQL-string to replace the appropriate
    ? with NULLs. We will assume there are no non-parameter ?-arguments --
    but to avoid confusion, we try to raise an exception if we suspect
    there are any.

    Arguments:
        sql_str (str): The SQL string to clean
        args (tuple[any]): The arguments to clean

    Returns:
        cleaned_sql_str (str): The SQL string with null parameters replaced
        cleaned_args (tuple[any]): The cleaned arguments with nulls replaced
    """
    if None not in args:
        return (sql_str, args)

    quote_char = None
    is_escaped = False

    result = []
    result_args = []
    current_start_index = 0
    next_arg_index = 0

    for i, c in enumerate(sql_str):
        if c == "?":
            if is_escaped:
                raise ValueError(
                    f"{sql_str=} appears to have an escaped ? - this is not supported with None-arguments"
                )

            if quote_char is not None:
                raise ValueError(
                    f"{sql_str=} appears to have a quoted ? - this is not supported with None-arguments"
                )

            if next_arg_index >= len(args):
                raise ValueError(
                    f"{sql_str=} has a ? without a matching argument (args={args})"
                )

            if args[next_arg_index] is None:
                result.append(sql_str[current_start_index:i])
                result.append("NULL")
                current_start_index = i + 1
            else:
                result_args.append(args[next_arg_index])

            next_arg_index += 1
            continue

        if is_escaped:
            is_escaped = False
            continue

        if c == "\\":
            is_escaped = True
            continue

        if quote_char is not None:
            if c == quote_char:
                quote_char = None
            continue

        if c == "'" or c == '"':
            quote_char = c

    if next_arg_index != len(args):
        raise ValueError(f"{sql_str=} has an argument without a matching ? ({args=})")

    result.append(sql_str[current_start_index:])
    return "".join(result), tuple(result_args)



from typing import Any, Literal, Optional, Tuple, TYPE_CHECKING
from rqdb.errors import DBError
from rqdb.logging import log
from rqdb.result import BulkResult, ResultItem, ResultItemCursor
from rqdb.preprocessing import get_sql_command, clean_nulls
import time
import secrets

if TYPE_CHECKING:
    from rqdb.connection import Connection


class Cursor:
    """A synchronous cursor for executing queries on a rqlite cluster."""

    def __init__(
        self,
        connection: "Connection",
        read_consistency: Literal["none", "weak", "strong"],
        freshness: str,
    ):
        self.connection = connection
        """The underlying connection to the rqlite cluster."""

        self.read_consistency = read_consistency
        """The read consistency to use when executing queries."""

        self.freshness = freshness
        """The freshness to use when executing none read consistency queries."""

        self.cursor: Optional[ResultItemCursor] = None
        """If we have a cursor object we are passing through to, this is it."""

        self.rows_affected: Optional[int] = None
        """The number of rows affected by the last query."""

        self.last_insert_id: Optional[int] = None
        """The last insert id after the last query"""

    @property
    def rowcount(self) -> int:
        """Returns the number of rows in the result set."""
        if self.cursor is None:
            return 0
        return self.cursor.rowcount

    def execute(
        self,
        operation: str,
        parameters: Optional[tuple] = None,
        raise_on_error: bool = True,
        read_consistency: Optional[Literal["none", "weak", "strong"]] = None,
        freshness: Optional[str] = None,
    ) -> ResultItem:
        """Executes a single query and returns the result. This will also
        update this object so that fetchone() and related functions can be used
        to fetch the results instead.

        Args:
            operation (str): The query to execute.
            parameters (tuple): The parameters to pass to the query.
            raise_on_error (bool): If True, raise an error if the query fails. If
                False, you can check the result item's error property to see if
                the query failed.
            read_consistency (Optional[Literal["none", "weak", "strong"]]):
                The read consistency to use when executing the query. If None,
                use the default read consistency for this cursor.
            freshness (Optional[str]): The freshness to use when executing
                none read consistency queries. If None, use the default freshness
                for this cursor.

        Returns:
            ResultItem: The result of the query.
        """
        if parameters is None:
            parameters = tuple()
        if read_consistency is None:
            read_consistency = self.read_consistency
        if freshness is None:
            freshness = self.freshness

        self.cursor = None
        self.rows_affected = None
        self.last_insert_id = None

        command = get_sql_command(operation)
        cleaned_query, parameters = clean_nulls(operation, parameters)
        is_read = command == "SELECT"
        request_id = secrets.token_hex(4)

        if is_read:

            def msg_supplier(max_length: Optional[int]) -> str:
                abridged_query = operation
                if max_length is not None and len(abridged_query) > max_length:
                    abridged_query = abridged_query[:max_length] + "..."

                abridged_parameters = repr(parameters)
                if max_length is not None and len(abridged_parameters) > max_length:
                    abridged_parameters = abridged_parameters[:max_length] + "..."

                freshness_str = ""
                if read_consistency == "none":
                    freshness_str = f", {freshness=}"

                return f"  [RQLITE {command} @ {read_consistency}{freshness_str} {{{request_id}}}] - {repr(abridged_query)}; {abridged_parameters}"

            log(self.connection.log_config.read_start, msg_supplier)

            path = f"/db/query?level={read_consistency}"
            if read_consistency == "none":
                path += f"&freshness={freshness}"

            request_started_at = time.perf_counter()
            response = self.connection.fetch_response(
                "POST",
                path,
                json=[[cleaned_query, *parameters]],
                headers={"Content-Type": "application/json; charset=UTF-8"},
            )
        else:

            def msg_supplier(max_length: Optional[int]) -> str:
                abridged_query = operation
                if max_length is not None and len(abridged_query) > max_length:
                    abridged_query = abridged_query[:max_length] + "..."

                abridged_parameters = repr(parameters)
                if max_length is not None and len(abridged_parameters) > max_length:
                    abridged_parameters = abridged_parameters[:max_length] + "..."

                return f"  [RQLITE {command} {{{request_id}}}] - {repr(abridged_query)}; {abridged_parameters}"

            log(self.connection.log_config.write_start, msg_supplier)

            request_started_at = time.perf_counter()
            response = self.connection.fetch_response(
                "POST",
                "/db/execute",
                json=[[cleaned_query, *parameters]],
                headers={"Content-Type": "application/json; charset=UTF-8"},
            )

        payload = response.json()
        request_time = time.perf_counter() - request_started_at

        def msg_supplier(max_length: Optional[int]) -> str:
            abridged_payload = response.text
            if max_length is not None and len(abridged_payload) > max_length:
                abridged_payload = abridged_payload[:max_length] + "..."

            return f"    {{{request_id}}} in {request_time:.3f}s -> {repr(abridged_payload)}"

        log(
            self.connection.log_config.read_response
            if is_read
            else self.connection.log_config.write_response,
            msg_supplier,
        )

        if "error" in payload:
            error = DBError(
                f'error outside of result: {payload["error"]}', payload["error"]
            )
            if error.is_stale:
                assert read_consistency == "none"

                def msg_supplier(max_length: Optional[int]) -> str:
                    return f"    {{{request_id}}} ->> stale read, retrying with weak consistency"

                log(self.connection.log_config.read_stale, msg_supplier)
                return self.execute(
                    operation,
                    parameters,
                    raise_on_error=raise_on_error,
                    read_consistency="weak",
                    freshness=freshness,
                )
            raise error

        results = payload.get("results")
        result = ResultItem.parse(results[0]) if results else ResultItem()
        if raise_on_error:
            result.raise_on_error(f"{request_id=}; {command=}; {parameters=}")

        if is_read and result.rowcount > 0:
            self.cursor = result.cursor()

        self.rows_affected = result.rows_affected
        self.last_insert_id = result.last_insert_id

        return result

    def executemany2(
        self,
        operations: Tuple[str],
        seq_of_parameters: Optional[Tuple[Tuple[Any]]] = None,
        transaction: bool = True,
        raise_on_error: bool = True,
    ) -> BulkResult:
        """Executes multiple operations within a single request and, by default, within
        a transaction.

        Unlike the standard DB-API executemany(), this method accepts different
        operations and parameters for each operation.

        Regardless of what type of operations are passed in, they will executed
        as if they are updates, i.e., no result rows will be returned.

        Args:
            operations (Tuple[str]): The operations to execute.
            seq_of_parameters (Tuple[Tuple[Any]]): The parameters to pass to each operation.
            transaction (bool): If True, execute the operations within a transaction.
            raise_on_error (bool): If True, raise an error if any of the operations fail. If
                False, you can check the result item's error property to see if the
                operation failed.

        Returns:
            BulkResult: The result of the query.

        Raises:
            ValueError: If the number of operations and parameters do not match.
        """
        if seq_of_parameters is None:
            seq_of_parameters = tuple(tuple() for _ in operations)

        if len(operations) != len(seq_of_parameters):
            raise ValueError(
                f"Number of operations ({len(operations)}) does not match number of parameters ({len(seq_of_parameters)})"
            )

        path = "/db/execute"
        if transaction:
            path += "?transaction"

        cleaned_request = []
        for operation, parameters in zip(operations, seq_of_parameters):
            cleaned_query, parameters = clean_nulls(operation, parameters)
            cleaned_request.append([cleaned_query, *parameters])

        request_id = secrets.token_hex(4)

        def msg_supplier(max_length: Optional[int]) -> str:
            abridged_request = repr(cleaned_request)
            if max_length is not None and len(abridged_request) > max_length:
                abridged_request = abridged_request[:max_length] + "..."

            return f"  [RQLITE BULK {path} {{{request_id}}}] - {abridged_request}"

        log(self.connection.log_config.write_start, msg_supplier)

        request_started_at = time.perf_counter()
        response = self.connection.fetch_response(
            "POST",
            path,
            json=cleaned_request,
            headers={"Content-Type": "application/json; charset=UTF-8"},
        )
        payload = response.json()
        request_time = time.perf_counter() - request_started_at

        def msg_supplier(max_length: Optional[int]) -> str:
            abridged_payload = response.text
            if max_length is not None and len(abridged_payload) > max_length:
                abridged_payload = abridged_payload[:max_length] + "..."

            return f"    {{{request_id}}} in {request_time:.3f}s -> {abridged_payload}"

        log(
            self.connection.log_config.write_response,
            msg_supplier,
        )

        result = BulkResult.parse(payload)
        if raise_on_error:
            result.raise_on_error(f"{request_id=}; {operations=}; {seq_of_parameters=}")

        return result

    def executemany3(
        self,
        operation_and_parameters: Tuple[Tuple[str, Tuple[Any]]],
        transaction: bool = True,
        raise_on_error: bool = True,
    ) -> BulkResult:
        """Executes multiple operations within a single request and, by default, within
        a transaction.

        Unlike the standard DB-API executemany(), this method accepts different
        operations and parameters for each operation.

        Regardless of what type of operations are passed in, they will executed
        as if they are updates, i.e., no result rows will be returned.

        Args:
            operations_and_parameters (Tuple[Tuple[str, Tuple[Any]]]):
                The operations and corresponding parameters to execute.
            transaction (bool): If True, execute the operations within a transaction.
            raise_on_error (bool): If True, raise an error if any of the operations fail. If
                False, you can check the result item's error property to see if the
                operation failed.

        Returns:
            BulkResult: The result of the query.

        Raises:
            ValueError: If the number of operations and parameters do not match.
        """
        path = "/db/execute"
        if transaction:
            path += "?transaction"

        cleaned_request = []
        for operation, parameters in operation_and_parameters:
            cleaned_query, parameters = clean_nulls(operation, parameters)
            cleaned_request.append([cleaned_query, *parameters])

        request_id = secrets.token_hex(4)

        def msg_supplier(max_length: Optional[int]) -> str:
            abridged_request = repr(cleaned_request)
            if max_length is not None and len(abridged_request) > max_length:
                abridged_request = abridged_request[:max_length] + "..."

            return f"  [RQLITE BULK {path} {{{request_id}}}] - {abridged_request}"

        log(self.connection.log_config.write_start, msg_supplier)

        request_started_at = time.perf_counter()
        response = self.connection.fetch_response(
            "POST",
            path,
            json=cleaned_request,
            headers={"Content-Type": "application/json; charset=UTF-8"},
        )
        payload = response.json()
        request_time = time.perf_counter() - request_started_at

        def msg_supplier(max_length: Optional[int]) -> str:
            abridged_payload = response.text
            if max_length is not None and len(abridged_payload) > max_length:
                abridged_payload = abridged_payload[:max_length] + "..."

            return f"    {{{request_id}}} in {request_time:.3f}s -> {abridged_payload}"

        log(
            self.connection.log_config.write_response,
            msg_supplier,
        )

        result = BulkResult.parse(payload)
        if raise_on_error:
            result.raise_on_error(f"{request_id=}; {operation_and_parameters=}")

        return result

    def fetchone(self) -> Optional[list]:
        """Fetches the next row from the cursor. If there are no more rows,
        returns None.
        """
        if self.cursor is None:
            return None
        result = self.cursor.fetchone()
        if result is None:
            self.cursor = None
        return result

    def fetchall(self) -> list:
        """Fetches all rows from the cursor. If there are no more rows,
        returns an empty list.
        """
        if self.cursor is None:
            return []
        result = self.cursor.fetchall()
        self.cursor = None
        return result

    def fetchmany(self, size: int) -> list:
        """Fetches the next `size` rows from the cursor. If there are no more rows,
        returns an empty list.
        """
        if self.cursor is None:
            return []
        result = self.cursor.fetchmany(size)
        if len(result) < size:
            self.cursor = None
        return result

    def close(self):
        """No-op"""
        pass



from typing import List, Literal, Optional, Tuple, Union, TYPE_CHECKING
from rqdb.errors import (
    ConnectError,
    MaxAttemptsError,
    MaxRedirectsError,
    UnexpectedResponse,
)
import rqdb.logging
import random
import aiohttp
from rqdb.async_cursor import AsyncCursor
import sys
import io
import inspect
import secrets
import time


class AsyncConnection:
    """Describes an asynchronous description of a pool of nodes which
    make up a rqlite cluster. There is no actual connection in this
    class, however the name is chosed to signify its similarity to
    a Connection in a DBAPI 2.0 style api.
    """

    def __init__(
        self,
        hosts: List[str],
        timeout: int = 5,
        max_redirects: int = 2,
        max_attempts_per_host: int = 2,
        read_consistency: Literal["none", "weak", "strong"] = "weak",
        freshness: str = "5m",
        log: Union[rqdb.logging.LogConfig, bool] = True,
    ):
        """Initializes a new synchronous Connection. This is typically
        called with the alias rqlite.connect

        Args:
            hosts (list[str]): A list of hostnames or ip addresses where
                the rqlite cluster is running. Each address can be a
                host:port pair - if the port is omitted, the default
                port of 4001 is used. If the host is specified in ipv6
                format, then it should be surrounded in square brackets
                if the port is specified, i.e., [::1]:4001.

            timeout (int): The number of seconds to wait for a response
                from a host before giving up.

            max_redirects (int): The maximum number of redirects to follow
                when executing a query.

            max_attempts_per_host (int): The maximum number of attempts to
                make to a given host before giving up.

            read_consistency (str): The read consistency to use for cursors
                by default.

            freshness (str): Default freshness value for none-consistency reads.
                Guarrantees the response is no more stale than the given duration,
                specified as a golang Duration string, e.g., "5m" for 5 minutes

            log (bool or LogConfig): If True, logs will have the default settings.
                If False, logs will be disabled. If a LogConfig, the
                configuration of the logs.
        """
        log_config: rqdb.logging.LogConfig = None
        if log is True:
            log_config = rqdb.logging.LogConfig()
        elif log is False:
            log_config = rqdb.logging.DISABLED_LOG_CONFIG
        else:
            log_config = log

        self.hosts: List[Tuple[str, int]] = [parse_host(h) for h in hosts]
        """The host addresses to connect to as a list of (hostname, port) pairs."""

        self.timeout: int = timeout
        """The number of seconds to wait for a response from a node."""

        self.max_redirects: int = max_redirects
        """The maximum number of redirects to follow when executing a query."""

        self.max_attempts_per_host: int = max_attempts_per_host
        """The maximum number of attempts to make to a node before giving up. We
        will attempt all nodes in a random order this many times per node before
        giving up.
        """

        self.read_consistency: Literal["none", "weak", "strong"] = read_consistency
        """The default read consistency when initializing cursors."""

        self.freshness: str = freshness
        """Default freshness value for none-consistency reads."""

        self.log_config = log_config
        """The log configuration for this connection and all cursors it creates."""

        self.session: aiohttp.ClientSession = None
        """The aiohttp session used to make requests to the cluster."""

    async def __aenter__(self) -> "AsyncConnection":
        """Initializes the client session"""
        self.session = aiohttp.ClientSession()
        await self.session.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Closes the client session"""
        session = self.session
        self.session = None
        await session.__aexit__(exc_type, exc_val, exc_tb)

    def cursor(
        self,
        read_consistency: Literal["none", "weak", "strong"] = None,
        freshness: str = None,
    ) -> "AsyncCursor":
        """Creates a new cursor for this connection.

        Args:
            read_consistency: The read consistency to use for this cursor. If
                None, the default read consistency for this connection will be
                used.
            freshness: The freshness value to use for this cursor. If None,
                the default freshness value for this connection will be used.

        Returns:
            A new cursor for this connection.
        """
        if read_consistency is None:
            read_consistency = self.read_consistency
        if freshness is None:
            freshness = self.freshness
        return AsyncCursor(self, read_consistency, freshness)

    async def fetch_response(
        self,
        method: Literal["GET", "POST"],
        uri: str,
        json: dict = None,
        headers: dict = None,
    ) -> aiohttp.ClientResponse:
        """Fetches a response from the server by requesting it from a random node. If
        a connection error occurs, this method will retry the request on a different
        node until it succeeds or the maximum number of attempts is reached.

        Args:
            method (str): The HTTP method to use for the request.
            uri (str): The URI of the request.
            json (dict): The json data to send with the request.
            headers (dict): The headers to send with the request.

        Returns:
            aiohttp.ClientResponse: If a successful response is received, it is returned.
                It has already been __aenter__'d.

        Raises:
            MaxAttemptsError: If the maximum number of attempts is reached before
                we get a successful response.
            UnexpectedResponse: If one of the rqlite nodes returns a response
                we didn't expect
        """
        node_path: List[Tuple[str, Exception]] = []

        async def attempt_host(host: Tuple[str, int]) -> aiohttp.ClientResponse:
            try:
                return await self.fetch_response_with_host(
                    *host,
                    method,
                    uri,
                    json,
                    headers,
                )
            except ConnectError as e:

                def msg_supplier(max_length: Optional[int]) -> str:
                    str_error = str(e)
                    if max_length is not None and len(str_error) > max_length:
                        str_error = str_error[:max_length] + "..."

                    return f"Failed to connect to node {e.host} - {str_error}"

                rqdb.logging.log(
                    self.log_config.connect_timeout, msg_supplier, exc_info=True
                )
                node_path.append((e.host, e))
            except MaxRedirectsError as e:

                def msg_supplier(max_length: Optional[int]) -> str:
                    str_redirect_path = str(e.redirect_path)
                    if max_length is not None and len(str_redirect_path) > max_length:
                        str_redirect_path = str_redirect_path[:max_length] + "..."

                    return f"Max redirects reached for node {e.host} (redirect path: {str_redirect_path})"

                rqdb.logging.log(
                    self.log_config.non_ok_response, msg_supplier, exc_info=True
                )
                node_path.append((e.host, e))
            except UnexpectedResponse as e:

                def msg_supplier(max_length: Optional[int]) -> str:
                    str_error = str(e)
                    if max_length is not None and len(str_error) > max_length:
                        str_error = str_error[:max_length] + "..."

                    return f"Unexpected response from node {e.host}: {str_error}"

                rqdb.logging.log(
                    self.log_config.non_ok_response, msg_supplier, exc_info=True
                )
                raise

        first_node_idx = random.randrange(0, len(self.hosts))
        if resp := await attempt_host(self.hosts[first_node_idx]):
            return resp

        remaining_nodes = [
            h for idx, h in enumerate(self.hosts) if idx != first_node_idx
        ]
        random.shuffle(remaining_nodes)
        node_ordering = [self.hosts[first_node_idx]] + remaining_nodes
        index = 0
        attempt = 1

        while index < len(self.hosts) or attempt < self.max_attempts_per_host:
            if index + 1 < len(self.hosts):
                index += 1
            else:
                index = 0
                attempt += 1

            if resp := await attempt_host(node_ordering[index]):
                return resp

        raise MaxAttemptsError(node_path)

    async def fetch_response_with_host(
        self,
        host: str,
        port: int,
        method: Literal["GET", "POST"],
        uri: str,
        json: dict = None,
        headers={},
    ) -> aiohttp.ClientResponse:
        """Fetches a response from a particular host, and returns the status
        code and headers. The response body is written to the given destination.

        This will follow up to the maximum number of redirects.

        Args:
            host (str): The host to fetch the response from.
            port (int): The port to use when connecting to the host.
            method (str): The HTTP method to use.
            uri (str): The URI to fetch.
            json (dict): The JSON to send in the request body.
            headers (dict): The headers to send in the request in addition to the content type.
                The dictionary should be accessed only using lowercase keys.

        Returns:
            aiohttp.ClientResponse: The response from the server.

        Raises:
            ConnectTimeout: If the connection times out.
            MaxRedirectsError: If the maximum number of redirects is exceeded.
            UnexpectedResponse: If the server returns a response we didn't expect.
        """
        redirect_path = []
        original_host = f"http://{host}:{port}{uri}"
        current_host = original_host
        response = None
        while len(redirect_path) < self.max_redirects:
            try:
                response = await self.session.request(
                    method,
                    current_host,
                    json=json,
                    headers=headers,
                    timeout=aiohttp.ClientTimeout(
                        connect=self.timeout,
                        sock_read=self.timeout,
                        total=self.timeout * 10,
                    ),
                    allow_redirects=False,
                )
                await response.__aenter__()

                if response.status in (301, 302, 303, 307, 308):
                    redirected_to = response.headers["Location"]
                    current_host = redirected_to
                    redirect_path.append(redirected_to)
                    await response.__aexit__(None, None, None)
                    continue

                if response.status < 200 or response.status > 299:
                    await response.__aexit__(None, None, None)
                    raise UnexpectedResponse(
                        current_host,
                        f"Unexpected response from {current_host}: {response.status} {response.reason}",
                    )

                return response
            except aiohttp.ClientConnectionError:
                if response is not None:
                    await response.__aexit__(*sys.exc_info())
                raise ConnectError(f"Connection to {host}:{port} failed", current_host)

        raise MaxRedirectsError(original_host, redirect_path)

    async def backup(self, file: io.BytesIO, /, raw: bool = False) -> None:
        """Backup the database to a file.

        Args:
            file (io.BytesIO): The file to write the backup to. May be an asyncio
                stream, i.e., file.write may be a coroutine function.
            raw (bool): If true, the backup will be in raw SQL format. If false,
                the backup will be in the smaller sqlite format
        """
        path = "/db/backup"
        if raw:
            path += "?fmt=sql"

        request_id = secrets.token_hex(4)
        rqdb.logging.log(
            self.log_config.backup_start,
            lambda _: f"  [RQLITE BACKUP {{{request_id}}}] raw={raw}",
        )
        request_started_at = time.perf_counter()
        resp = await self.fetch_response("GET", path)
        is_coroutine = inspect.iscoroutinefunction(file.write)
        try:
            while True:
                chunk = await resp.content.read(1024 * 4)
                if not chunk:
                    break
                if is_coroutine:
                    await file.write(chunk)
                else:
                    file.write(chunk)
        finally:
            await resp.__aexit__(None, None, None)

        time_taken = time.perf_counter() - request_started_at
        rqdb.logging.log(
            self.log_config.backup_end,
            lambda _: f"    {{{request_id}}} in {time_taken:.3f}s ->> backup fully written",
        )


def parse_host(host: str) -> Tuple[str, int]:
    """Parses a host:port pair into an ip address and port.

    Args:
        host (str): A host:port pair, or just a host

    Returns:
        A tuple of (ip, port)
    """
    if not host:
        raise ValueError("host must not be empty")

    num_colons = host.count(":")
    if num_colons > 1:
        # ipv6; must be of the form [host]:port if port is specified
        if host[0] != "[":
            return host, 4001

        close_square_bracket_idx = host.find("]")
        return host[1:close_square_bracket_idx], int(
            host[close_square_bracket_idx + 1 :]
        )

    if num_colons == 0:
        return host, 4001

    hostname, port_str = host.split(":")
    return hostname, int(port_str)



from typing import Any, Callable, Optional, TypedDict
import dataclasses
import logging


class LogMessageConfig(TypedDict):
    """Configures a single log message within rqlite."""

    enabled: bool
    """True if the message should be logged, False otherwise. If not
    present, assumed to be True
    """

    method: Callable[[str], Any]
    """The function to call to log the message. If not present,
    then this will be set based on the level of the message.
    For example, a level of DEBUG implies that the method is
    effectively logging.debug.

    The method should support "exc_info=True" as a keyword argument.
    """

    level: int
    """The level of the message. If not present, assumed to be
    logging.DEBUG. The level does not have to be one of the default
    logging levels - the method will be a partial variant of logging.log
    with the level as the first argument.

    The level is ignored if the method is set.
    """

    max_length: Optional[int]
    """The approximate maximum length of the message. This may be
    implemented differently depending on which message is being
    configured. If not present assumed to be None, for no maximum
    length.
    """


@dataclasses.dataclass(frozen=True)
class LogConfig:
    """Describes the configuration of rqlite's logging."""

    read_start: LogMessageConfig = dataclasses.field(
        default_factory=lambda: LogMessageConfig(enabled=True, level=logging.DEBUG)
    )
    """Configures the message to log when cursor.execute
    is called with a SELECT query.
    """

    read_response: LogMessageConfig = dataclasses.field(
        default_factory=lambda: LogMessageConfig(enabled=True, level=logging.DEBUG)
    )
    """Configures the message to log when we get the response
    from the server for a SELECT query.
    """

    read_stale: LogMessageConfig = dataclasses.field(
        default_factory=lambda: LogMessageConfig(enabled=True, level=logging.DEBUG)
    )
    """Configures the message to log when we get a response from
    the server for a SELECT query, but the response indicates we
    must retry because the data is not sufficiently fresh. This
    occurs only on reads with the read consistency level "none".
    """

    write_start: LogMessageConfig = dataclasses.field(
        default_factory=lambda: LogMessageConfig(enabled=True, level=logging.DEBUG)
    )
    """Configures the message to log when cursor.execute
    is called with a non-SELECT query.
    """

    write_response: LogMessageConfig = dataclasses.field(
        default_factory=lambda: LogMessageConfig(enabled=True, level=logging.DEBUG)
    )
    """Configures the message to log when we get the response
    from the server for a non-SELECT query.
    """

    connect_timeout: LogMessageConfig = dataclasses.field(
        default_factory=lambda: LogMessageConfig(enabled=True, level=logging.WARNING)
    )
    """Configures the message to log when a connection attempt
    to one of the host nodes fails.
    """

    hosts_exhausted: LogMessageConfig = dataclasses.field(
        default_factory=lambda: LogMessageConfig(enabled=True, level=logging.CRITICAL)
    )
    """Configures the message to log when we are going to give
    up on a given query because we have exhausted all attempts on
    all nodes. This implies the cluster is unresponsive or we cannot
    reach the cluster.
    """

    non_ok_response: LogMessageConfig = dataclasses.field(
        default_factory=lambda: LogMessageConfig(enabled=True, level=logging.WARNING)
    )
    """Configures the message to log when we get a response from
    the server that is not OK or is a redirect when one is not
    expected, such as when we have exceeded the maximum number of
    redirects.
    """

    backup_start: LogMessageConfig = dataclasses.field(
        default_factory=lambda: LogMessageConfig(enabled=True, level=logging.INFO)
    )
    """Configures the message to log when we start attempting a backup."""

    backup_end: LogMessageConfig = dataclasses.field(
        default_factory=lambda: LogMessageConfig(enabled=True, level=logging.INFO)
    )
    """Configures the message to log when we finish attempting a backup."""


DISABLED_LOG_CONFIG = LogConfig(
    read_start=LogMessageConfig(enabled=False),
    read_response=LogMessageConfig(enabled=False),
    read_stale=LogMessageConfig(enabled=False),
    write_start=LogMessageConfig(enabled=False),
    write_response=LogMessageConfig(enabled=False),
    connect_timeout=LogMessageConfig(enabled=False),
    hosts_exhausted=LogMessageConfig(enabled=False),
    non_ok_response=LogMessageConfig(enabled=False),
    backup_start=LogMessageConfig(enabled=False),
    backup_end=LogMessageConfig(enabled=False),
)
"""The log configuration which disables all logging."""


def log(
    config: LogMessageConfig,
    msg_supplier: Callable[[Optional[int]], str],
    exc_info: bool = False,
) -> None:
    """Logs a message if the config is enabled.

    Args:
        config: The configuration of the message to log.
        msg_supplier: A function which returns the message to log.
            Passed the approximate length of the message if there is
            one.
        exc_info: True to pass the current exception to the logger,
            False not to.
    """
    if not config.get("enabled", True):
        return

    max_length = config.get("max_length", None)
    message = msg_supplier(max_length)

    method = config.get("method", None)
    if method is not None:
        if not exc_info:
            method(message)
        else:
            method(message, exc_info=True)
        return

    level = config.get("level", logging.DEBUG)
    logging.log(level, message, exc_info=exc_info)
